/*
** dyn.c for elfsh
** 
** Started on  Fri Nov  2 15:17:36 2001 mayhem
** Last update Sat Mar 15 07:35:08 2003 mayhem
*/
#include "elfsh.h"


/* Provide human readable output for .dynamic entries */
void		vm_dynentinfo(elfshobj_t	*file, 
			      Elf32_Dyn		*entry, 
			      char		*info)
{
  char		*str;

  switch (entry->d_tag)
    {
    case DT_NEEDED:
    case DT_SONAME:
    case DT_RPATH:
      str = elfsh_get_dynentry_string(file, entry);
      snprintf(info, 99, "%s", str);
      break;
    case DT_SYMBOLIC:
      sprintf(info, "Current object first checked during symbol resolving");
      break;
    case DT_TEXTREL:
      sprintf(info, "Relocation can overwrite .text");
      break;
    case DT_BIND_NOW:
      sprintf(info, "Symbol binding is now completed before execution");
      break;
    case DT_PLTREL:
      sprintf(info, "%d", entry->d_un.d_val);
      break;
    case DT_NULL:
      sprintf(info, "END OF SECTION");
      break;
    case DT_PLTRELSZ:
    case DT_RELASZ:
    case DT_RELAENT:
    case DT_STRSZ:
    case DT_RELSZ:
    case DT_RELENT:
    case DT_SYMENT:
      sprintf(info, "%u bytes", entry->d_un.d_val);
      break;
    case DT_PLTGOT:
    case DT_STRTAB:
    case DT_SYMTAB:
    case DT_HASH:
    case DT_RELA:
    case DT_INIT:
    case DT_FINI:
    case DT_REL:
    case DT_DEBUG:
    case DT_JMPREL:
      sprintf(info, "0x%08X", (u_int) entry->d_un.d_ptr);
      break;
    default:
      sprintf(info, "[?]");
      break;
    }
}




/* Display the .dynamic section entries */
int		cmd_dyn()
{
  Elf32_Dyn	*actual;
  int		num;
  int		index;
  int		typenum;
  char		*type;
  char		*type_short;
  regex_t	*tmp;
  char		buff[256];
  char		type_unk[ELFSH_MEANING + 1];
  char		info[256];

  /* Fetch the section and init the regex */
  if ((actual = elfsh_get_dynamic(world.current, &num)) == NULL)
    RET(-1);
  CHOOSE_REGX(tmp);
  printf(" [SHT_DYNAMIC]\n [Object %s]\n\n", world.current->name);

  /* Loop on the section */
  for (index = 0; index < num && actual[index].d_tag != DT_NULL; index++)
    {
      
      /* Get the data to be printed */
      typenum = elfsh_get_dynentry_type(actual + index);
      if (typenum >= ELFSH_DYNAMIC_MAX)
	type = type_short = vm_build_unknown(type_unk, "UNK", typenum);
      else
	{
	  type       = (char *) elfsh_dynentry_type[typenum].desc;
	  type_short = (char *) elfsh_dynentry_type[typenum].name;
	}

      vm_dynentinfo(world.current, actual + index, info);

      /* Print if the regex match */
      snprintf(buff, sizeof(buff), " [%02u] %-33s => %19s {%s}", 
	       index, type, info, type_short);

      if (!tmp || (tmp && !regexec(tmp, buff, 0, 0, 0)))
	puts(buff);

    }
  
  puts("");
  return (0);
}







