/*
** findrel.c for elfsh
**
** Started on  Sat Feb 22 17:06:17 2003 mayhem
** Last update Tue Mar 18 04:24:02 2003 mayhem
*/
#include "elfsh.h"




/* Resolve a virtual address */
char		*vm_reverse(elfshobj_t *file, u_int vaddr)
{
  char		*str;
  int		off;
  char		*new;

  str = elfsh_reverse_symbol(file, vaddr, &off);
  if (str == NULL)
    str = elfsh_reverse_dynsymbol(file, vaddr, &off);
  if (str == NULL)
    return (str);
  if (off)
    {
      XALLOC(new, strlen(str) + 20, NULL);
      sprintf(new, "<%s + %u>", str, off);
    }
  else
    new = strdup(str);
  return (new);
}



/* Find the missing relocation table for an ET_EXEC object */
int		cmd_findrel()
{
  elfshobj_t	*file;
  elfshsect_t	*cur;
  elfshsect_t	*dst;
  u_int		index;
  char		soff[30];
  char		doff[30];
  u_int		count;

  /* Sanity checks */
  file = world.current;
  if (file == NULL)
    RETERR("[elfsh:findrel] Invalid NULL parameter\n");
  if (elfsh_read_obj(file) < 0)
    RETERR("[elfsh:findrel] Cannot read object\n");

  printf(" [*] EXTRA relocs for %s \n\n",
	 file->name);

  /* For all sections of the current object */
  for (count = 0, cur = file->sectlist; cur; cur = cur->next)
    {

      /* Do not look for cross references on unmapped or void sections */
      if (cur->data == NULL)
	{
	  if (!world.state.vm_quiet)
	    printf(" [*] Passing %-20s [NO DATA]\n", cur->name);
	  continue;
	}
      if (cur->shdr->sh_addr == NULL)
	{
	  if (!world.state.vm_quiet)
	    printf(" [*] Passing %-20s [UNMAPPED]\n", cur->name);
	  continue;
	}
      
      /* Find the relocation entries, and print it */
      elfsh_find_rel(cur);
      if (cur->rel == NULL || cur->srcref == 0)
	{
	  if (!world.state.vm_quiet)
	    printf(" [*] Passing %-20s [NO RELOC ENTRY]\n", cur->name);
	  continue;
	}
     
      /* Print the relocs for this section */
      puts("");
      for (index = 0; index < cur->srcref; index++)
	{
	  
	  /* Print the offset buffers */
	  if (cur->rel[index].off_src)
	    snprintf(soff, sizeof(soff), "+ %u", cur->rel[index].off_src);
	  else
	    *soff = 0x00;
	  if (cur->rel[index].off_dst)
	    snprintf(doff, sizeof(doff), "+ %u", cur->rel[index].off_dst);
	  else
	    *doff = 0x00;
	  
	  /* Print to stdout the current rel ent */
	  dst = elfsh_get_section_by_index(file,
					   cur->rel[index].idx_dst, 
					   NULL, 
					   NULL);
	  printf(" [%02u] From %20s %s TO %20s %s [ %08X -> %08X ] \n",
		 index, cur->name, soff, dst->name, doff, 
		 cur->shdr->sh_addr + cur->rel[index].off_src,
		 dst->shdr->sh_addr + cur->rel[index].off_dst);
	  count++;
	}
    }
  
  /* Print final banner */
  printf("\n [*] Total number of absolute references : %u\n\n", count);
  for (cur = world.current->sectlist; cur; cur = cur->next)
    printf(" [*] Section %-20s srcref[%u] dstref[%u] \n", 
	   cur->name, cur->srcref, cur->dstref);
  puts("");

  return (0);
}

