/*
** dtors.c for libelfsh
**
** Started on  Tue Feb 26 22:12:38 2002 mayhem
** Last update Mon Mar 17 04:35:14 2003 mayhem
*/
#include "libelfsh.h"



/* Read the destructor array in .dtors */
void		*elfsh_get_dtors(elfshobj_t *file, int *num)
{
  elfshsect_t	*new;

  new = elfsh_get_section_by_name(file, ELFSH_SECTION_NAME_DTORS, NULL, NULL, NULL);
  if (!new)
    return (NULL);
  if (!new->data)
    {
      new->nbr = new->shdr->sh_size / 4;
      new->data = elfsh_load_section(file, new->shdr);
      if (!new->data)
	return (NULL);
      file->secthash[ELFSH_SECTION_DTORS] = new;
    }
  if (num)
    *num = new->nbr;
  return (new->data);
}






/* Modify a DTORS entry */
int		elfsh_set_dtors_entry_by_index(elfshobj_t *file, int index, int new_addr)
{
  int		nbr;
  int		*dtors;

  dtors = elfsh_get_dtors(file, &nbr);
  if (!dtors)
    return (-1);
  if (index >= file->secthash[ELFSH_SECTION_DTORS]->nbr)
    ELFSH_SETERROR("libelfsh: index too big . \n", -1);
  dtors[index] = new_addr;
  return (0);
}





/* Modify a DTORS entry using its name */
int		elfsh_set_dtors_entry_by_name(elfshobj_t *file, char *name, int new_addr)
{
  int		nbr;
  int		*dtors;
  Elf32_Sym	*sym;
  int		index;

  dtors = elfsh_get_dtors(file, &nbr);
  if (!dtors)
    return (-1);

  sym = elfsh_get_symbol_by_name(file, name);
  if (!sym)
    return (-1);

  for (index = 0; index < nbr; index++)
    if (dtors[index] == sym->st_value)
      {
	dtors[index] = new_addr;
	return (0);
      }

  ELFSH_SETERROR("libelfsh: Cant find dtors entry for this symbol . \n", -1);
}




/* Return a entry giving its parent and its index : used as INTERNAL hash handler */
u_int		*elfsh_get_dtors_entry_by_index(u_int *dtors, u_int index)
{
  return (dtors + index);
}


/* Return a entry giving its parent and its index : used as INTERNAL hash handler */
u_int		elfsh_set_dtors_entry(u_int *dtors, u_int vaddr)
{
  if (!dtors)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", -1);
  *dtors = vaddr;
  return (0);
}


/* Return a entry value giving its parent and its index : used as INTERNAL hash handler */
u_int		elfsh_get_dtors_entry(u_int *dtors)
{
  if (!dtors)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", -1);
  return (*dtors);
}
