/*
** Dynsym.c for libelfsh
** 
** Started on  Mon Feb 26 04:13:29 2001 mayhem
** Last update Mon Mar 17 03:59:11 2003 mayhem
*/
#include "libelfsh.h"





/* Return the dynsymbol name giving its index in the dynamic symbol string table */
char		*elfsh_get_dynsymbol_name(elfshobj_t *file, Elf32_Sym *s)
{
  elfshsect_t	*sect;

  if (file->secthash[ELFSH_SECTION_DYNSYM] == NULL && 
      !elfsh_get_dynsymtab(file, NULL))
    return (NULL);

  /* Use the section string table if the symbol is a section */
  if (elfsh_get_symbol_type(s) == STT_SECTION)
    {
      sect = elfsh_get_section_from_sym(file, s);
      return (sect != NULL ? elfsh_get_section_name(file, sect) : NULL);
    }

  return ((char *) file->secthash[ELFSH_SECTION_DYNSTR]->data + s->st_name);
}



/* Return the used offset in .dynstr or -1 if failed */
int		elfsh_set_dynsymbol_name(elfshobj_t *file, Elf32_Sym *s, char *name)
{
  char		*str;
  u_int		len;
  u_int		new_len;

  if (file == NULL || s == NULL || name == NULL)
    ELFSH_SETERROR("[libelfsh:set_dynsymbol_name] Invalid NULL parameter", -1);

  if (file->secthash[ELFSH_SECTION_DYNSYM] == NULL && 
      elfsh_get_dynsymtab(file, NULL) == NULL)
    ELFSH_SETERROR("[libelfsh:get_dynsymbol_name] Cannot retreive symbol table",
		   -1);

  /* Else use the symbol string table */
  if (file->secthash[ELFSH_SECTION_DYNSTR] == NULL || 
      file->secthash[ELFSH_SECTION_DYNSTR]->data == NULL)
    return (NULL);

  /* Change the name */
  str = file->secthash[ELFSH_SECTION_DYNSTR]->data;
  str += s->st_name;
  len = strlen(str);
  new_len = strlen(name);

  /* Do not allocate new place if possible */
  if (len >= new_len)
    strcpy(str, name);
  /* Append the name to .strtab */
  else
    s->st_name = elfsh_insert_in_dynstr(file, name);

  return (s->st_name);
}






/* Return a ptr on the dynamic symbol table num is filled with the entries total number */
void		*elfsh_get_dynsymtab(elfshobj_t *file, int *num)
{
  elfshsect_t	*new;
  int		strindex;
  int		nbr;

  if (file == NULL)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", NULL);


  /* Load the .dynsym */
  if (file->secthash[ELFSH_SECTION_DYNSYM] == NULL)
    {
      new = elfsh_get_section_by_type(file, SHT_DYNSYM, NULL, &strindex, &nbr, 0);
      if (new == NULL)
	return (NULL);
      new->nbr = nbr / ELFSH_SYMTAB_ENTRY_SIZE;
      new->data = elfsh_load_section(file, new->shdr);
      if (new->data == NULL) 
	return (NULL);
      file->secthash[ELFSH_SECTION_DYNSYM] = new;

      /* Now retreive the dynamic symbol string table .dynstr */
      new = elfsh_get_section_by_index(file, strindex, NULL, &nbr);
      if (new == NULL)
	return (NULL);
      new->data = elfsh_load_section(file, new->shdr);
      if (new->data == NULL)
	return (NULL);
      new->nbr = nbr;
      file->secthash[ELFSH_SECTION_DYNSTR] = new;

      /* Sort the table */
      elfsh_sync_sorted_symtab(file->secthash[ELFSH_SECTION_DYNSYM]);
    }

  if (num != NULL)
    *num = file->secthash[ELFSH_SECTION_DYNSYM]->nbr;
  return (file->secthash[ELFSH_SECTION_DYNSYM]->data);
}







/* Return the dynamic symbol name giving its value */
char		*elfsh_reverse_dynsymbol(elfshobj_t *file, u_int value, int *offset)
{
  Elf32_Sym	*sorted;
  int		num;
  int		index;
  char		*str;

  /* To shut gcc up with -Wall */
  str = NULL;

  /* for .dtors_END and .ctors_end */
  if (value == 0 || value == 0xFFFFFFFF)
    return (NULL);  
  if (file == NULL)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", NULL);

  if (elfsh_get_dynsymtab(file, &num) == NULL)
    return (NULL);

  sorted = file->secthash[ELFSH_SECTION_DYNSYM]->altdata;
  for (index = 0; index < num; index++)
    if (sorted[index].st_value >= (u_int) value)
      {
	if (!index)
	  {
	    str = NULL;
	    break;
	  }
	*offset = (int) value - sorted[index - 1].st_value;
	str = elfsh_get_dynsymbol_name(file, sorted + index - 1);
	break;
      }
  return (str);
}





/* Return the symbol entry giving its name */
Elf32_Sym	*elfsh_get_dynsymbol_by_name(elfshobj_t *file, char *name)
{
  Elf32_Sym	*ret;
  int		index;
  int		size = 0;
  char		*actual;

  if (file == NULL)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", NULL);
  if (NULL == elfsh_get_dynsymtab(file, &size))
    return (NULL);
  if (NULL == name)
    return (NULL);

  ret = file->secthash[ELFSH_SECTION_DYNSYM]->data;
  for (index = 0; index < size; index++)
    {
      actual = elfsh_get_dynsymbol_name(file, ret + index);
      if (actual && !strcmp(actual, name))
	return (ret + index);
    }
  
  ELFSH_SETERROR("[libelfsh] Symbol not found", NULL);
}


























