/*
** fixup.c for elfsh
** 
** Started on  Fri Jul 27 04:56:06 2001 mayhem
** Last update Mon Mar 17 04:04:36 2003 mayhem
*/
#include "libelfsh.h"


/* Put the bss physically in the file */
int		elfsh_fixup_bss(elfshobj_t *file)
{
  elfshsect_t	*bss;
  elfshsect_t	*actual;

  /* Find the bss or return an error */
  bss = elfsh_get_section_by_name(file, ELFSH_SECTION_NAME_BSS, NULL, NULL, NULL);
  if (bss == NULL)
    ELFSH_SETERROR("[libelfsh] Cant find .bss section \n", -1);
  if (bss->shdr->sh_size == 0) 
    ELFSH_SETERROR("[libelfsh] .bss section size == 0 \n", -1);

  /* Add the bss data in the file in order to be able to insert sections after it */
  if (bss->data == NULL)
    {
      XALLOC(bss->data, bss->shdr->sh_size, -1);

      /* Extend the segment for .bss */
      bss->phdr->p_filesz += bss->shdr->sh_size;

      /* Update the file offset for the section header table if necessary */
      if (file->hdr->e_shoff > bss->shdr->sh_offset)
	file->hdr->e_shoff += bss->shdr->sh_size;

      /* Update the file offset for all next sections */
      for (actual = bss->next; actual; actual = actual->next)
	actual->shdr->sh_offset += bss->shdr->sh_size;
    }

  return (0);
}



/* TERMINAL FUNCTION ! Fixup size for section symbols, if symtab doesnt exist, create it */
elfshsect_t	*elfsh_fixup_symtab(elfshobj_t *file, int *strindex)
{
  elfshsect_t	*symtab;
  elfshsect_t	*list;
  Elf32_Shdr	hdr;
  Elf32_Sym	new;
  Elf32_Sym	*actual;
  u_int		index;

  if (file == NULL)
    ELFSH_SETERROR("[libelfsh:fixup_symtab] Invalid NULL parameter\n", NULL);

  /* Create the table if it does not exist */
  if (file->secthash[ELFSH_SECTION_SYMTAB] == NULL)
    {
      XALLOC(symtab, sizeof (elfshsect_t), NULL);
      hdr = elfsh_create_section_header(0, SHT_SYMTAB, 0, 0, 0, 0, 0, 0, 0, 0);
      symtab->name = strdup(ELFSH_SECTION_NAME_SYMTAB);
      elfsh_insert_unmapped_section(file, symtab, hdr, NULL);
      file->secthash[ELFSH_SECTION_SYMTAB] = symtab;
    }
  else
    symtab = file->secthash[ELFSH_SECTION_SYMTAB];

  /* Walk to the sht and fixup the symbols in .symtab, or create them if unexistant */
  for (list = file->sectlist; list != NULL; list = list->next)
    {
      actual = elfsh_get_sym_from_shtentry(file, list->shdr);
      if (actual == NULL)
	{
	  new = elfsh_create_symbol(list->shdr->sh_addr, 
				    list->shdr->sh_size, 
				    STT_SECTION, 0, 0);
	  elfsh_insert_symbol(symtab, &new, list->name);
	}
      else
	actual->st_size = list->shdr->sh_size;
    }

  /* Fill strindex */
  if (strindex != NULL)
    elfsh_get_section_by_name(file, ELFSH_SECTION_NAME_STRTAB, strindex, NULL, NULL);

  /* Fixup all 0 lenght non-section-typed symbols */
  actual = symtab->data;
  for (index = 0; index < symtab->nbr; index++, actual++)
    if (actual->st_value && !actual->st_size && index + 1 < symtab->nbr)
      actual->st_size = actual[1].st_value - actual->st_value;

  elfsh_sync_sorted_symtab(symtab);
  return (symtab);
}



/* Rebuild symbol string table if unexistant */
elfshsect_t	*elfsh_rebuild_strtab(elfshobj_t *file)
{
  elfshsect_t	*strtab;
  Elf32_Shdr	hdr;

  if (file == NULL)
    return (NULL);

  /* Create the table if it does not exist */
  if (file->secthash[ELFSH_SECTION_STRTAB])
    {
      XALLOC(strtab, sizeof (elfshsect_t), NULL);
      hdr = elfsh_create_section_header(0, SHT_STRTAB, 0, 0, 0, 0, 0, 0, 0, 0);
      strtab->name = strdup(ELFSH_SECTION_NAME_STRTAB);
      elfsh_insert_unmapped_section(file, strtab, hdr, NULL);
      file->secthash[ELFSH_SECTION_STRTAB] = strtab;
    }
  return (file->secthash[ELFSH_SECTION_STRTAB]);
}
