/*
** got.c for elfsh
**
** Started on  Sun Jun 24 21:30:41 2001 mayhem
** Last update Mon Mar 17 04:07:07 2003 mayhem
*/
#include "libelfsh.h"





/* Read the global offset table in section .got */
void		*elfsh_get_got(elfshobj_t *file, int *num)
{
  elfshsect_t	*new;

  new = elfsh_get_section_by_name(file, ELFSH_SECTION_NAME_GOT, NULL, NULL, NULL);
  if (new == NULL)
    return (NULL);
  if (new->data == NULL)
    {
      new->nbr = new->shdr->sh_size / 4;
      new->data = elfsh_load_section(file, new->shdr);
      if (new->data == NULL)
	return (NULL);
      file->secthash[ELFSH_SECTION_GOT] = new;
    }

  if (num != NULL)
    *num = new->nbr;
  return (new->data);
}






/* Modify a GOT entry */
int		elfsh_set_got_entry_by_index(elfshobj_t *file, int index, int new_addr)
{
  int		nbr;
  int		*got;

  got = elfsh_get_got(file, &nbr);
  if (!got)
    return (-1);
  if (index >= nbr)
    ELFSH_SETERROR("[libelfsh] Index too big", -1);
  got[index] = new_addr;
  return (0);
}




/* Modify the got entry for the dynamic symbol 'name' */
int		elfsh_set_got_entry_by_name(elfshobj_t *file, char *name, int new_addr)
{
  Elf32_Sym	*sym;
  int		foffset;
  int		got;

  if (file == NULL || name == NULL)
    ELFSH_SETERROR("[libelfsh:set_got_entry_by_name] Invalid NULL parameter\n",
		   -1);

  sym = elfsh_get_dynsymbol_by_name(file, name);
  if (sym == NULL)
    ELFSH_SETERROR("[libelfsh] Unknown dynamic symbol", 0);

  if (!elfsh_is_pltentry(file, sym))
    ELFSH_SETERROR("[libelfsh] Symbol is not a Procedure Linkage Table entry", 0);

  if (elfsh_get_got(file, NULL) == NULL)
    ELFSH_SETERROR("[libelfsh] Cant retreive Global offset Table", -1);

  /* We pick up the got entry address from the plt */
  foffset = elfsh_get_foffset_from_vaddr(file, sym->st_value);
  elfsh_raw_read(file, foffset + 2, &got, sizeof(got));
  foffset = elfsh_get_foffset_from_vaddr(file, got);
  elfsh_raw_write(file, foffset, &new_addr, sizeof(got));
  return (0);
}



/* Return a entry giving its parent and its index */
u_int		*elfsh_get_got_entry_by_index(u_int *got, u_int index)
{
  return (got + index);
}


/* Return a entry giving its parent and its index : used as INTERNAL hash handler */
u_int		elfsh_set_got_entry(u_int *got, u_int vaddr)
{
  if (!got)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", -1);
  *got = vaddr;
  return (0);
}

/* Return a entry value giving its parent and its index : used as INTERNAL hash handler */
u_int		elfsh_get_got_entry(u_int *got)
{
  if (!got)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", -1);
  return (*got);
}
