/*
** reloc.c for elfsh
** 
** Started on  Tue Feb 27 14:34:50 2001 mayhem
** Last update Tue Mar 18 05:45:31 2003 mayhem
*/
#include "libelfsh.h"




/* Return the relocation type */
u_int	elfsh_get_reltype(Elf32_Rel *r)
{
  if (!r)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", -1);
  return (ELF32_R_TYPE(r->r_info));
}

/* Return the relocation type */
u_int	elfsh_set_reltype(Elf32_Rel *r, Elf32_Word type)
{
  if (!r)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", -1);
  r->r_info = (r->r_info & 0xFF00) + ELF32_R_TYPE(type);
  return (0);
}

/* Return the relocation symbol index */
u_int	elfsh_get_relsym(Elf32_Rel *r)
{
  if (!r)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", -1);
  return (ELF32_R_SYM(r->r_info));
}

/* Change the relocation symbol index */
u_int	elfsh_set_relsym(Elf32_Rel *r, Elf32_Word sym)
{
  if (!r)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", -1);
  r->r_info = (r->r_info & 0x00FF) + ((sym << 8) & 0xFF00);
  return (0);
}


/* Return the relocation offset */
Elf32_Addr	elfsh_get_reloffset(Elf32_Rel *r)
{
  if (!r)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", -1);
  return (r->r_offset);
}

/* Change the relocation offset for this entry */
int		elfsh_set_reloffset(Elf32_Rel *r, Elf32_Addr off)
{
  if (!r)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter", -1);
  r->r_offset = off;
  return (0);
}




/* Return the 'range'th relocation table */
elfshsect_t	*elfsh_get_reloc(elfshobj_t *file, u_int range, u_int *nbr)
{
  elfshsect_t	*s;
  int		index;
  int		size;

  s = elfsh_get_section_by_type(file, SHT_REL, &index, NULL, &size, range);
  if (!s)
    return (NULL);
  if (!s->data)
    {
      s->data = elfsh_load_section(file, s->shdr);
      s->nbr  = s->shdr->sh_size / sizeof(Elf32_Rel);
    }
  if (nbr)
    *nbr = s->nbr;
  return (s);
}


/* Return the symbol name associated with the relocation entry */
Elf32_Sym	*elfsh_get_symbol_from_reloc(elfshobj_t *file, Elf32_Rel *r)
{
  int		tmp;
  char		ismod;
  
  if (!file || (!file->hdr && !elfsh_get_hdr(file)))
    return (NULL);
  
  tmp = ELF32_R_SYM(r->r_info);
  ismod = (file->hdr->e_type == ET_REL ? 1 : 0);
  
  if (ismod)
    {
      if (NULL == file->secthash[ELFSH_SECTION_SYMTAB] && 
	  NULL == elfsh_get_symtab(file, NULL))
	return (NULL);
      return ((Elf32_Sym *) file->secthash[ELFSH_SECTION_SYMTAB]->data + tmp);
    }
  else
    {
      if (NULL == file->secthash[ELFSH_SECTION_DYNSYM] && 
	  NULL == elfsh_get_dynsymtab(file, NULL))
	return (NULL);
      return ((Elf32_Sym *) file->secthash[ELFSH_SECTION_DYNSYM]->data + tmp);
    }

}


/* Used as internal handler for elfsh hashes */
Elf32_Rel	*elfsh_get_relent_by_index(Elf32_Rel *table, u_int index)
{
  if (!table)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter\n", NULL);
  return (table + index);
}




/* Create relocation table for section */
elfshrel_t	*elfsh_find_rel(elfshsect_t *sect)
{
  elfshsect_t	*target;
  elfshrel_t	*rel;
  int		*dword;
  char		*str;
  u_int		vaddr;
  u_int		index;

  /* Sanity checks */
  if (sect == NULL)
    ELFSH_SETERROR("[libelfsh:find_rel] Invalid NULL parameter\n", NULL);
  else if (sect->data == NULL)
    ELFSH_SETERROR("[libelfsh:find_rel] Section empty\n", NULL);
  else if (sect->shdr->sh_addr == NULL)
    ELFSH_SETERROR("[libelfsh:find_rel] Section unmapped\n", NULL);
  else if (sect->rel)
    return (sect->rel);
  
  /* These sections must not be relocated, but passed to relative */
  if (sect->shdr->sh_type == SHT_DYNSYM             ||
      /*      sect->shdr->sh_type == SHT_HASH               || */
      sect->shdr->sh_type == SHT_REL                || 
      sect->shdr->sh_type == SHT_RELA               || 
      sect->shdr->sh_type == SHT_DYNAMIC            ||
      !strcmp(sect->name, ELFSH_SECTION_NAME_GOT)   ||
      !strcmp(sect->name, ELFSH_SECTION_NAME_CTORS) ||
      !strcmp(sect->name, ELFSH_SECTION_NAME_DTORS))
    return (NULL);

  /* Free existing rel[] entries if existing (the section data probably changed) */
  sect->srcref = sect->dstref = 0;

  /* Read the actual section and find valid references */
  str = sect->data;
  for (dword = (int *) str; 
       ((char *) dword + 4) <= ((char *) sect->data + sect->shdr->sh_size); 
       dword = (int *) str)
    {
      target = elfsh_isvalidptr(sect->parent, *dword);
      if (target != NULL)
	{
	  sect->srcref++;
	  target->dstref++;
	}
      str++;
    }
  
  /* allocate the array */
  if (sect->srcref == 0)
    ELFSH_SETERROR("[libelfsh:find_rel] No need to relocate section\n", NULL);
  XALLOC(rel, sect->srcref * sizeof(elfshrel_t), NULL);
  
  /* Read the actual section again and create section rel table */
  str = sect->data;
  for (index = 0, dword = (int *) str; 
       ((char *) dword + 4) <= ((char *) sect->data + sect->shdr->sh_size); 
       dword = (int *) str)
    {
      target = elfsh_isvalidptr(sect->parent, *dword);
      if (target != NULL)
	{
	  vaddr = sect->shdr->sh_addr + (u_int) str - (u_int) sect->data;
	  rel[index].idx_src = sect->index;
	  rel[index].off_src = vaddr - sect->shdr->sh_addr;
	  rel[index].idx_dst = target->index;
	  rel[index].off_dst = *dword - target->shdr->sh_addr;
	  index++;
	}
      str++;
    }

  /* Return the array */
  sect->rel = rel;
  return (sect->rel);
}


