/*
** sht.c for libelfsh
** 
** Started on  Mon Feb 26 04:16:18 2001 mayhem
** Last update Tue Mar 18 04:28:04 2003 mayhem
*/
#include "libelfsh.h"



/* Return a ptr on the section header table */
void		*elfsh_get_sht(elfshobj_t *file, int *num)
{
  elfshsect_t	*s;
  char		*name;

  if (file == NULL)
    ELFSH_SETERROR("[libelfsh:get_sht] Invalid NULL parameter\n", NULL);
  
  /* Load the section header table and allocate every elfshsect_t */
  if (file->sht != NULL)
    goto end;

  /* Load SHT and reconstruct it if necessary */
  else if (elfsh_load_sht(file) < 0 && elfsh_rebuild_sht(file) < 0)
    ELFSH_SETERROR("[libelfsh:get_sht] Cant load SHT\n", NULL);
  
  /* 
  ** Load the section string table if not already done in elfsh_rebuild_sht()
  ** If we cant get it, do not return an error, but continue without it
  */
  if (file->secthash[ELFSH_SECTION_SHSTRTAB] == NULL)
    {
      s = elfsh_get_section_by_index(file, file->hdr->e_shstrndx, NULL, NULL);
      if (s != NULL)
	{
	  file->secthash[ELFSH_SECTION_SHSTRTAB] = s;
	  if (s->data == NULL)
	    s->data = elfsh_load_section(file, s->shdr);
	  s->nbr = s->shdr->sh_size;
	}
    }

  /* 
  ** Now associate sections with section names . We could not have done 
  ** this in elfsh_load_sht() since .shstrtab was not loaded .
  */
  for (s = file->sectlist; s; s = s->next)
    {
      name = elfsh_get_section_name(file, s);
      s->name = (name != NULL ? strdup(name) : NULL);
    }
  
  /* Some final stuffs */
 end:
  if (num)
    *num = file->hdr->e_shnum;
  return (file->sht);
}







/* Return an array of section header */
int		elfsh_load_sht(elfshobj_t *file)
{
  elfshsect_t	*new;
  int		size;
  int		index;

  if (file == NULL)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter\n", -1);
  else if (file->hdr == NULL && NULL == elfsh_get_hdr(file))
    return (-1);
  else if (file->sht != NULL)
    return (0);
  
  size = file->hdr->e_shnum * file->hdr->e_shentsize;

  if (!file->hdr->e_shoff)
    ELFSH_SETERROR("[libelfsh] SHT file offset is NULL\n", -1);

  XSEEK(file->fd, file->hdr->e_shoff, SEEK_SET, -1);
  XALLOC(file->sht, size, -1);
  XREAD(file->fd, file->sht, size, -1);

  /* 
  ** Allocate a elfshsect_t per section and put it 
  ** in the file->sectlist double linked list 
  */
  for (index = 0; index < file->hdr->e_shnum; index++)
    {
      XALLOC(new, sizeof(elfshsect_t), -1);
      if (elfsh_add_section(file, 
			    new, index, NULL, ELFSH_SHIFTING_COMPLETE) < 0)
	return (-1);
    }

  return (0);
}







/* Retreive the section header giving the section symbol from .symtab */
Elf32_Shdr		*elfsh_get_shtentry_from_sym(elfshobj_t *file, 
						     Elf32_Sym	*sym)
{
  elfshsect_t		*sect;

  sect = elfsh_get_section_from_sym(file, sym);
  return (sect == NULL ? NULL : sect->shdr);
}





/* Retreive the symbol for the section giving the section header */
Elf32_Sym		*elfsh_get_sym_from_shtentry(elfshobj_t *file, 
						     Elf32_Shdr *hdr)
{
  Elf32_Sym		*current;
  int			index;

  if (file == NULL || hdr == NULL)
    ELFSH_SETERROR("[libelfsh:get_sym_from_shtentry] Invalid NULL parameter\n", 
		   NULL);
  else if (file->sht == NULL && NULL == elfsh_get_sht(file, NULL)) 
    return (NULL);
  else if (file->secthash[ELFSH_SECTION_SYMTAB] == NULL && 
	   NULL == elfsh_get_symtab(file, NULL))
    return (NULL);
  else if (!file->secthash[ELFSH_SECTION_SYMTAB]->nbr)
    ELFSH_SETERROR("[libelfsh:get_sym_from_shtentry] No symbol table entry\n",
		   NULL);

  for (index = 0, current = file->secthash[ELFSH_SECTION_SYMTAB]->data; 
       index < file->secthash[ELFSH_SECTION_SYMTAB]->nbr; 
       index++)
    if (elfsh_get_symbol_type(current + index) == STT_SECTION && 
	current[index].st_value == hdr->sh_addr)
      return (current + index);

  ELFSH_SETERROR("[libelfsh:get_sym_from_shtentry] No symbol for this section",
		 NULL);
}







/* Create a section header */
Elf32_Shdr		elfsh_create_section_header(Elf32_Word name, 
						    Elf32_Word type, 
						    Elf32_Word flags,
						    Elf32_Addr addr, 
						    Elf32_Off offset, 
						    Elf32_Word size,
						    Elf32_Word link, 
						    Elf32_Word info,
						    Elf32_Word align,
						    Elf32_Word entsize)
{
  Elf32_Shdr		new;

  new.sh_name = name;
  new.sh_type = type;
  new.sh_flags = flags;
  new.sh_addr = addr;
  new.sh_offset = offset;
  new.sh_size = size;
  new.sh_link = link;
  new.sh_info = info;
  new.sh_addralign = align;
  new.sh_entsize = entsize;
  return (new);
}




/* Add a section header to the object */
int		elfsh_insert_section_header(elfshobj_t	*file, 
					    Elf32_Shdr	hdr, 
					    u_int	range, 
					    char	*name)
{
  elfshsect_t	*tmp;
  int		index;
  int		shname;

  /* Preliminary checks */
  if (file == NULL || 
      (file->hdr == NULL && elfsh_get_hdr(file) == NULL) || 
      (file->sht == NULL && elfsh_get_sht(file, NULL) == NULL))
    return (-1);

  /* Fixup the range */
  if (range == ELFSH_SECTION_LAST)
    range = file->hdr->e_shnum;
  else if (range > file->hdr->e_shnum)
    ELFSH_SETERROR("[libelfsh] Invalid range for injection\n", -1);

  /* Insert the name string in .shstrtab */
  shname = elfsh_insert_in_shstrtab(file, name);
  if (shname < 0)
    ELFSH_SETERROR("[libelfsh] Cannot inject data in .shstrtab\n", -1);
  hdr.sh_name = (u_int) shname;
  if (range > file->secthash[ELFSH_SECTION_SHSTRTAB]->index)
    hdr.sh_offset += strlen(name) + 1;

  /* Extend the SHT and fill the new slot */
  XREALLOC(file->sht, file->sht, 
	   file->hdr->e_shentsize * (file->hdr->e_shnum + 1), -1);

  if (range != file->hdr->e_shnum)
    memmove(file->sht + range + 1, 
	    file->sht + range, 
	    (file->hdr->e_shnum - range) * file->hdr->e_shentsize);

  file->sht[range] = hdr;
  file->hdr->e_shnum++;

  /* Update foffset for all section header whoose section is placed after SHT */
  for (tmp = file->sectlist, index = 0; index < file->hdr->e_shnum; index++)
    {

      /* 
      ** Since we reallocated the sht, we have to update the shdr pointer
      ** since its pointing to the old chunk !
      ** Pass headers without body (this include the just inserted one) 
      */
      if (index != range && tmp)
	{
	  tmp->shdr = file->sht + index;
	  tmp = tmp->next;
	}
      
      /* Now update the file offset for the actual section */
      if (file->sht[index].sh_offset >= file->hdr->e_shoff)
	file->sht[index].sh_offset += file->hdr->e_shentsize;

    }

  /* If we have an influence on the sht file offset, update it */
  if (hdr.sh_offset < file->hdr->e_shoff)
    file->hdr->e_shoff += hdr.sh_size;
  
  /* Update the .shstrtab index in the ELF header if necessary */
  if (range <= file->hdr->e_shstrndx)
    file->hdr->e_shstrndx++;

  return (range);
}




/* Get the section's name in .shstrtab */
char		*elfsh_get_section_name(elfshobj_t *file, 
					elfshsect_t *s)
{
  volatile char	*str;

  /* Sanity check */
  if (file == NULL || s == NULL || 
      file->secthash[ELFSH_SECTION_SHSTRTAB] == NULL ||
      (file->sht == NULL && NULL == elfsh_get_sht(file, NULL)))
    return (NULL);
  
  /* Return section's name */
  str = file->secthash[ELFSH_SECTION_SHSTRTAB]->data;
  return ((char *) str + s->shdr->sh_name);
}


/* Set the section's name in .shstrtab */
int			elfsh_set_section_name(elfshobj_t	*file, 
					       elfshsect_t	*s, 
					       char		*name)
{
  char		*str;
  u_int		len;
  u_int		new_len;

  /* Sanity check */
  if (file == NULL || s == NULL || name == NULL || 
      file->secthash[ELFSH_SECTION_SHSTRTAB] == NULL ||
      (file->sht == NULL && elfsh_get_sht(file, NULL) == NULL))
    return (NULL);
  
  /* Return section's name */
  str = file->secthash[ELFSH_SECTION_SHSTRTAB]->data;
  str += s->shdr->sh_name;
  len = strlen(str);
  new_len = strlen(name);

  /* Do not allocate new place if possible */
  if (len >= new_len)
    strcpy(str, name);
  /* Append the name to .shstrtab */
  else
    s->shdr->sh_name = elfsh_insert_in_shstrtab(file, name);

  return (s->shdr->sh_name);
}




/* Tell elfsh that we dont want the sht in the output file */
void		elfsh_remove_sht(elfshobj_t *file)
{
  file->shtrm = 1;
}




/* SET/GET the allocatable flag */
char	elfsh_get_section_allocflag(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return ((s->sh_flags & SHF_ALLOC));
}

char	elfsh_set_section_allocflag(Elf32_Shdr *s, char f)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  if (f)
    s->sh_flags |= SHF_ALLOC;
  else
    s->sh_flags &= (~SHF_ALLOC);
  return (0);
}


/* SET/GET the writable flag */
char	elfsh_get_section_writableflag(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return ((s->sh_flags & SHF_WRITE));
}

char	elfsh_set_section_writableflag(Elf32_Shdr *s, char f)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  if (f)
    s->sh_flags |= SHF_WRITE;
  else
    s->sh_flags &= (~SHF_WRITE);
  return (0);
}



/* SET/GET the executable flag */
char	elfsh_get_section_execflag(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return ((s->sh_flags & SHF_EXECINSTR));
}

char	elfsh_set_section_execflag(Elf32_Shdr *s, char f)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  if (f)
    s->sh_flags |= SHF_EXECINSTR;
  else
    s->sh_flags &= (~SHF_EXECINSTR);
  return (0);
}


/* SET/GET the mergeable flag */
char	elfsh_get_section_mergeableflag(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return ((s->sh_flags & SHF_MERGE));
}

char	elfsh_set_section_mergeableflag(Elf32_Shdr *s, char f)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  if (f)
    s->sh_flags |= SHF_MERGE;
  else
    s->sh_flags &= (~SHF_MERGE);
  return (0);
}


/* SET/GET the string flag */
char	elfsh_get_section_strflag(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return ((s->sh_flags & SHF_STRINGS));
}

char	elfsh_set_section_strflag(Elf32_Shdr *s, char f)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  if (f)
    s->sh_flags |= SHF_STRINGS;
  else
    s->sh_flags &= (~SHF_STRINGS);
  return (0);
}


/* SET/GET the link flag */
char	elfsh_get_section_linkflag(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return ((s->sh_flags & SHF_INFO_LINK));
}

char	elfsh_set_section_linkflag(Elf32_Shdr *s, char f)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  if (f)
    s->sh_flags |= SHF_INFO_LINK;
  else
    s->sh_flags &= (~SHF_INFO_LINK);
  return (0);
}


/* SET/GET the order flag */
char	elfsh_get_section_orderflag(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return ((s->sh_flags & SHF_LINK_ORDER));
}

char	elfsh_set_section_orderflag(Elf32_Shdr *s, char f)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  if (f)
    s->sh_flags |= SHF_LINK_ORDER;
  else
    s->sh_flags &= (~SHF_LINK_ORDER);
  return (0);
}





/* All the next functions are dumb read/write field access routine */
Elf32_Word		elfsh_get_section_info(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return (s->sh_info);
}

int		elfsh_set_section_info(Elf32_Shdr *s, Elf32_Word info)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  s->sh_info = info;
  return (0);
}

Elf32_Word		elfsh_get_section_entsize(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return (s->sh_entsize);
}

int		elfsh_set_section_entsize(Elf32_Shdr *s, Elf32_Word entsize)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  s->sh_entsize = entsize;
  return (0);
}

Elf32_Word	elfsh_get_section_link(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return (s->sh_link);
}

int		elfsh_set_section_link(Elf32_Shdr *s, Elf32_Word link)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  s->sh_link = link;
  return (0);
}




Elf32_Off	elfsh_get_section_foffset(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return (s->sh_offset);
}

int	elfsh_set_section_foffset(Elf32_Shdr *s, Elf32_Off offset)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  s->sh_offset = offset;
  return (0);
}

Elf32_Addr	elfsh_get_section_addr(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return (s->sh_addr);
}

int	elfsh_set_section_addr(Elf32_Shdr *s, Elf32_Addr addr)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  s->sh_addr = addr;
  return (0);
}


Elf32_Word	elfsh_get_section_align(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return (s->sh_addralign);
}

int	elfsh_set_section_align(Elf32_Shdr *s, Elf32_Word align)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  s->sh_addralign = align;
  return (0);
}

Elf32_Word	elfsh_get_section_size(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return (s->sh_size);
}

int	elfsh_set_section_size(Elf32_Shdr *s, Elf32_Word size)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  s->sh_size = size;
  return (0);
}

Elf32_Word	elfsh_get_section_type(Elf32_Shdr *s)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  return (s->sh_type);
}

int	elfsh_set_section_type(Elf32_Shdr *s, Elf32_Word type)
{
  if (!s)
    ELFSH_SETERROR("[libelfsh] Invalid NULL parameter \n", -1);
  s->sh_type = type;
  return (0);
}


/* This is usually not used except in the ELFsh hash handlers */
Elf32_Shdr		*elfsh_get_sht_entry_by_index(Elf32_Shdr *s, u_int index)
{
  return (s + index);
}








